/*
 * manager-resolvconf.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2025 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "manager-private.h"
#include "network-inador-public.h"
#include "network-inador-manager.h"

void _manager_resolvconf_handle_feed (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int name_len, wanted;
	unsigned char iface_prog[256];
	unsigned char iface_name[256], prog[256];
	char *point;
	void *resolv_entries;
	int entries_count;
	
	int ns_count;
	int search_count;
	int g, pos, ns_pos;
	int family;
	uint32_t iface_index;
	Interface *iface;
	union {
		struct in_addr v4;
		struct in6_addr v6;
	} addr;
	
	if (buffer_len < 3) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RESOLVCONF_FEED);
		return;
	}
	
	name_len = buffer[2];
	
	if (buffer_len < 3 + name_len || name_len == 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RESOLVCONF_FEED);
		return;
	}
	
	memcpy (iface_prog, &buffer[3], name_len);
	iface_prog[name_len] = 0;
	
	if (iface_prog[0] == '.') {
		/* No permitimos interfaces vacías */
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_RESOLVCONF_FEED);
		
		return;
	}
	
	point = strchr (iface_prog, '.');
	if (point == NULL) {
		prog[0] = 0;
		strncpy (iface_name, iface_prog, sizeof (iface_name));
	} else {
		point[0] = 0;
		strncpy (iface_name, iface_prog, sizeof (iface_name));
		strncpy (prog, &point[1], sizeof (prog));
	}
	
	iface = network_inador_get_iface_by_name (manager_client->manager->handle, iface_name);
	if (iface == NULL) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_RESOLVCONF_FEED);
		return;
	}
	iface_index = network_inador_iface_get_index (iface);
	
	pos = 3 + name_len;
	if (buffer_len < pos) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RESOLVCONF_FEED);
		return;
	}
	
	/* Revisar la cantidad de nameservers */
	ns_count = buffer[pos];
	pos++;
	
	/* TODO: Sumar las otras entradas */
	entries_count = ns_count;
	
	ns_pos = pos;
	/* Los nameserver vienen con un byte de familia y luego los bytes correspondientes a la IP */
	for (g = 0; g < ns_count; g++) {
		/* Revisar que en la longitud venga el byte de la familia */
		if (buffer_len < pos + 1) {
			_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RESOLVCONF_FEED);
			return;
		}
		
		family = buffer[pos];
		pos++; /* El byte de la familia */
		
		if (family != AF_INET && family != AF_INET6) {
			_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_FAMILY, NET_INADOR_COMMAND_RESOLVCONF_FEED);
			
			return;
		}
		
		if (family == AF_INET) {
			wanted = 4;
		} else if (family == AF_INET6) {
			wanted = 16;
		}
		
		if (buffer_len < pos + wanted) {
			_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RESOLVCONF_FEED);
			return;
		}
		pos += wanted;
	}
	
	/* TODO: Validar otro tipo de entradas */
	
	/* Notificar al cliente */
	_manager_send_executed (manager_client);
	
	/* Re-agrupar los datos para enviarlos al resolv_manager */
	pos = ns_pos;
	resolv_entries = NULL;
	for (g = 0; g < ns_count; g++) {
		family = buffer[pos];
		pos++; /* El byte de la familia */
		
		wanted = (family == AF_INET) ? 4 : 16;
		memcpy (&addr, &buffer[pos], wanted);
		pos += wanted;
		
		resolv_entries = network_inador_resolvconf_append_ns_entry (resolv_entries, family, &addr, iface_index, prog);
	}
	
	resolv_manager_process_resolvconf_entries (manager_client->manager->handle, resolv_entries);
	network_inador_resolvconf_free_all_entries (resolv_entries);
}

void _manager_resolvconf_handle_remove_feed (ManagerClientInfo *manager_client, unsigned char *buffer, int buffer_len) {
	int name_len;
	unsigned char iface_prog[256];
	unsigned char iface_name[256], prog[256];
	char *point;
	Interface *iface;
	uint32_t iface_index;
	
	name_len = buffer[2];
	
	if (buffer_len < 3 + name_len || name_len == 0) {
		_manager_send_error (manager_client, NET_INADOR_ERROR_INCOMPLETE_REQUEST, NET_INADOR_COMMAND_RESOLVCONF_REMOVE);
		return;
	}
	
	memcpy (iface_prog, &buffer[3], name_len);
	iface_prog[name_len] = 0;
	
	if (iface_prog[0] == '.') {
		/* No permitimos interfaces vacías */
		_manager_send_error (manager_client, NET_INADOR_ERROR_INVALID_VALUE, NET_INADOR_COMMAND_RESOLVCONF_REMOVE);
		
		return;
	}
	
	point = strchr (iface_prog, '.');
	if (point == NULL) {
		prog[0] = 0;
		strncpy (iface_name, iface_prog, sizeof (iface_name));
	} else {
		point[0] = 0;
		strncpy (iface_name, iface_prog, sizeof (iface_name));
		strncpy (prog, &point[1], sizeof (prog));
	}
	
	iface = network_inador_get_iface_by_name (manager_client->manager->handle, iface_name);
	if (iface == NULL) {
		iface_index = 0;
	} else {
		iface_index = network_inador_iface_get_index (iface);
	}
	
	resolv_manager_clear_entries_by_prog (manager_client->manager->handle, iface_index, prog);
	
	_manager_send_executed (manager_client);
}

